
Konzepte der Programmierung                       Universität Konstanz
Programmierkurs 2                Informatik & Informationswissenschaft
Sommer 2016                                Marc Scholl, Stefan Klinger

======================================================================

                             11. Juli 2016

----------------------------------------------------------------------

Dieses Kapitel ist nicht relevant für die Klausur!

----------------------------------------------------------------------

> module Main where
> 
> import Text.Parsec
> import Data.List ( foldl' )  -- cf. `pk2-11.lhs`

----------------------------------------------------------------------
Parsec — Combinator library for parser construction [1]


Um mit Parsec [1] spielen zu können müssen Sie die entsprechende
Haskell-Bibliothek [2] installieren.

    $ cabal install -j parsec

Der verwendete Paket-Manager `cabal` [3] ist auf `titan07` leider
nicht installiert, aber bei der Haskell-Platform [4] mit dabei.  Viele
Linux-Distributionen bieten das auch als Paket `cabal-install` an,
z.B. [5].


Abgesehen vom originalen Paper [1] gibt es zur zeit leider nicht viel
brauchbare Tutorial-Style Dokumentation zu Parsec, eine Liste [6] ist
bei GitHub.


----------------------------------------------------------------------
Fangen wir mal an...


Idee: Ein Parser liest möglichst viel von einer Eingabe (String von
      stdin oder aus Datei), und produziert ein Ergebnis vom Typ α.

Parsec einen Typ zur Verfügung der solche Parser beschreibt:

    type Parsec s u a

Das Typsystem von Parsec ist etwas komplex, wir verwenden einen
einfacheren Typ:

> type Parser a = Parsec String () a

Dinge vom Typ `Parser α` sollen Parser sein, die einen String lesen,
und einen Wert vom Typ α produzieren.

  | Die ganze Wahrheit?
  | 
  |     type Parsec s u a = ParsecT s u Identity a
  | 
  | `ParsecT s u m a` is a parser with stream type `s`, user state
  | type `u`, underlying monad `m` and return type `a`.  —siehe [7]
  | 
  | Das brauchen wir alles nicht!
  |
  | Wenn der GHCi `ParsecT s u m α` sagt, denken wir einfach `Parser α`.


Was wir brauchen ist eine Funktion die einen Parser auf eine Eingabe
anwendet.  Vordefiniert ist

    parse :: Parser α    -- ein Parser, den müssen wir bauen

          -> SourceName  -- Ein Name für die Quelle (typischerweise der
                            Dateiname aus dem gelesen wurde) wird für die
                            Ausgabe von Fehlermeldungen benötigt.
                         
          -> String      -- die Eingabe die gelesen werden soll
          
          -> Either ParseError α   -- ein Wert :: α, oder eine
                                      Fehlermeldung

Damit ich nicht immer einen Dateinamen eintippen muss:

> test :: Parser a -> String -> Either ParseError a
> test p s = parse p "input" s

`test` nimmt also einen Parser `p` und eine Eingabe `s`, und gibt mir
das Ergebnis oder eine Fehlermeldung zurück.



Einfache Parser
---------------

    char :: Char -> Parser Char

Der Parser `char x` konsumiert genau den angegebenen Buchstaben, und
gibt ihn zurück.

    test (char 'h') "hans"
    test (char 'h') "wurst"

Beobachtung:

  * Wenn der Parser nicht bekommt was er erwartet, dann wird eine
    Fehlermeldung erzeugt.

  * Ein Parser konsumiert nur den Anfang der Eingabe (soweit möglich,
    cf. später), "ans" und "urst" spielten keine Rolle!



Etwas allgemeiner, und auch schon vordefiniert:

    satisfy :: (Char -> Bool) -> Parser Char

akzeptiert Buchstaben die das Prädikat erfüllen.

    test (satisfy (> 't')) "aber"
    test (satisfy (> 't')) "wrtschk"



Vordefiniert ist: `instance Monad Parser`, es gibt also einen Parser
der nichts tut (liest) und trotzdem ein Ergebnis liefert:

    test (return 23) ""
    test (return 23) "fsdfsd"



Weitere einfache Parser:

    anyChar :: Parser Char    -- akzeptiert jeden Character
    letter :: Parser Char     -- nur Buchstaben       
    digit :: Parser Char      -- nur Ziffern

Ausprobieren:

    test anyChar "1"
    test anyChar "a"
    test letter "1"
    test letter "a"
    test digit "1"
    test digit "a"



Weil `Parser` Instanz von `Monad` ist, ist es auch Instanz von
`Applicative`, und von `Functor`:

Bauen einen Parser der den *Wert* einer Ziffer liefert:

> digit' :: Parser Int
> digit' = f <$> digit      -- `f` auf *Ergebnis* des Parsers anwenden
>   where
>     f c = fromEnum c - fromEnum '0'

    test digit' "5"
    test digit' "lala"

    test digit' "50"  --warum?

Schade: der Parser `digit` konsumiert nur ein Zeichen!



Sequenz
-------

Erinnerung:

    class Applicative m => Monad m where
      ...
      (>>) :: m a -> m b -> m b

und

    instance Monad Parser 

Frage: Welchen Typ hat das entsprechende (>>) ?

    (>>) :: Parser a -> Parser b -> Parser b 


> abc :: Parser Char
> abc = char 'a' >> char 'b' >> char 'c'

    test abc "abc"

Was erwarten wir als Ergebnis?


Äquivalent:

> abc' :: Parser Char
> abc' = do char 'a'
>           char 'b'
>           char 'c'



Bauen einen Parser, der einen bestimmten String erwartet:

> string' :: String -> Parser ()

Idee?

> string' "" = return ()
> string' (x:xs) = char x >> string' xs


Wie können wir daraus einen Parser bauen der den String ausserdem
zurückgibt?

> string'' :: String -> Parser String
> string'' xs = string' xs >> return xs

Schon vordefiniert ist

    string :: String -> Parser String



Alternative
-----------

    (<|>) :: Parser a -> Parser a -> Parser a

> xy :: Parser Char
> xy = char 'x' <|> char 'y'

    test xy "y"
    test xy "t"


Auch vordefiniert ist die Verallgemeinerung auf mehrere Alternativen:

    choice :: [Parser a] -> Parser a

> xyz :: Parser Char
> xyz = choice [ char 'x', char 'y', char 'z' ]

    test xyz "y"
    test xyz ""



Mit Sequenz und Alternative können wir schon interessante rekursive
Parser definieren.

Was tut das:

> bar :: Parser Int
> 
> bar = do char '('
>          inner <- bar
>          char ')'
>          after <- bar
>          return $ max (inner + 1) after
>       <|>
>       return 0


    test bar ""
    test bar "()(()(()))"



Predictive
----------

Obacht! `<|>` versucht den 2. Parser *nur dann*, wenn der erste
        überhaupt keinen Input konsumiert hat.

        (Grund: Performance — Parsec kann sofort entscheiden, welche
        der beiden Alternativen weiter verfolgt wird.)

> wrong :: Parser Char
> wrong = (char 'x' >> char 'y')
>         <|>
>         (char 'x' >> char 'z')

Was soll dieser Parser wohl tun?

    test wrong ""
    test wrong "xy"
    test wrong "xz"


Lösung: Gleiche linke Terme muss man *herausfaktorisieren*:

> right :: Parser Char
> right = char 'x' >> ( char 'y' <|> char 'z' )

    test right ""
    test right "xy"
    test right "xz"


Manchmal ist das Herausfaktorisieren nicht so einfach (oder
unmöglich), dann kann man sich mit dem `try`-Kmobinator behelfen.

    try :: Parser a -> Parser a

> ok :: Parser Char
> ok = try (char 'x' >> char 'y')
>      <|>
>      (char 'x' >> char 'z')

Wenn `p` fehlschlägt, dann tut `try p` so, als ob keine Eingabe
konsumiert wurde.  Deswegen versucht `<|>` die zweite Alternative!

    test ok ""
    test ok "xy"
    test ok "xz"

Obacht! `try p` muss sich die Ganze Eingabe merken, bis `p` definitiv
        erfolgreich ist, oder fehlschlägt!



Wiederholungen
--------------

    many :: Parser a -> Parser [a]     -- beliebig viele, oder…
    many1 :: Parser a -> Parser [a]    -- …mindestens eine Wiederholung.

Ausprobieren:

    test (many1 digit') ""
    test (many1 digit') "12345kjfdlksdf"

Das hatten wir schon einmal:

    foldl' (\x d-> 10 * x + d) 0 [1,2,3,4,5]

Wir müssen das nur auf das *Ergebnis eines Parsers* anwenden.  Wie?

> num :: Parser Int
> num = foldl' (\x d-> 10 * x + d) 0 <$> many1 digit'

    test num ""
    test num "123"
    test num "000123xxxx"

    test num "-123"

Wie könnte man noch mit dem `-` umgehen?

> int :: Parser Int
> int = do s <- (char '-' >> return negate)  -- konsumiere ein '-'
>               <|>                          -- oder 
>               return id                    -- konsumiere nichts
>          s <$> num         -- `s` ist jetzt `negate` oder `id`

    test int "-123"



Longest-Match Rule
------------------

Mal angenommen wir haben eine Sprache die unter anderem beliebig
benannte Variablen und die Konstante `pi` enthält.

> data Foo = Pi
>          | Var String
>          -- ...   
>          deriving Show

Foo soll eins von beiden sein

> fooBad :: Parser Foo
> fooBad = choice [ string "pi" >> return Pi
>                 , Var <$> many1 letter
>                 ]

    test fooBad "pi"
    test fooBad "x"

Gibt's ein Problem?

    test fooBad "pit"

Vordefiniert ist:

    notFollowedBy :: Show a => Parser a -> Parser ()

`notFollowedBy p` ist erfolgreich, wenn `p` fehlschlägt.  Der von `p`
konsumierte Input wird wieder hergestellt (ähnlich zu `try`).

> keyword :: String -> Parser ()
> keyword k = try $ do string k
>                      notFollowedBy alphaNum   -- !

> foo :: Parser Foo
> foo = choice [ keyword "pi" >> return Pi
>              , Var <$> many1 letter
>              ]

    test fooBad "pi"
    test fooBad "x"
    test fooBad "pit"


Übrignes:

    eof :: Parser ()
    eof = notFollowedBy anyChar

Damit können wir das Parsen der gesamten Eingabe erzwingen!

> entirely' :: Parser a -> Parser a
> entirely' p = do x <- p
>                  eof
>                  return x

    test (entirely' int) "123"
    test (entirely' int) "123x"



--- hier morgen weitermachen ---



Vordefiniert ist:

    between :: Parser a -> Parser b -> Parser c -> Parser c

> entirely'' :: Parser a -> Parser a
> entirely'' = between (return ()) eof



Listen mit Kommas oder Operatoren dazwischen
--------------------------------------------

    sepBy  :: Parser a -> Parser b -> Parser [a]
    sepBy1 :: Parser a -> Parser b -> Parser [a]

`sepBy p c` liest beliebig viele Wiederholungen von `p`, die alle
durch `c` getrennt sind.  `sepBy1` erwartet mindestens ein Auftauchen
von `p`.

> ints :: Parser [Int]
> ints = int `sepBy` char ','

    test ints ""
    test ints "324"
    test ints "324,454,1212"


Was wenn keine Kommas sondern Operatoren dazwischen stehen?

> sumOp :: Parser (Int -> Int -> Int)
> sumOp = choice [ char '+' >> return (+)
>                , char '-' >> return (-)
>                ]

> mulOp :: Parser (Int -> Int -> Int)
> mulOp = choice [ char '*' >> return (*)
>                , char '/' >> return div
>                ]

    chainl  :: Parser a -> Parser (a -> a -> a) -> a -> Parser a
    chainl1 :: Parser a -> Parser (a -> a -> a) -> Parser a

    chainr  :: Parser a -> Parser (a -> a -> a) -> a -> Parser a
    chainr1 :: Parser a -> Parser (a -> a -> a) -> Parser a

Die Parser `chainX p o z` parsen beliebig viele Wiederholungen von
`p`, getrennt durch `o`.  Die Ergebnisse der `p` werden durch links-
bzw. rechtsassoziative Anwendung der von `o` gelieferten Funktionen
zusammengesetz.  Bei 0 Wiederholungen wird `z` zurückgegeben.

Die Parser `chainX1 p o` erwarten mindestens ein Auftauchen von `p`.


    test (chainl1 int sumOp) "3-2-1"
    test (chainr1 int sumOp) "3-2-1"


> expr' :: Parser Int
> expr' = flip chainl1 sumOp
>         $
>         flip chainl1 mulOp
>         $
>         int
>         <|>
>         between (char '(') (char ')') expr'

    test expr' "10-(9-8)-7"
    test expr' "1+2*(3-4)"

    test expr' " 4 * (2 + 3) "


Lexeme
------

Ein “Lexem Parser” soll ein Parser sein, der ausserdem alle
nachfolgenden Leerzeichen konsumiert.

> lexeme :: Parser a -> Parser a
> lexeme p = do x <- p
>               spaces -- zero or more
>               return x

    test (entirely' $ lexeme int) "23     "


Tip: Veruchen Sie von Anfang an Lexeme-Parser zu bauen.

> parens :: Parser a -> Parser a
> parens = between (lexeme $ char '(') (lexeme $ char ')')

Eleganter wäre es natürlich `sumOp`, `mulOp` und `int` von Anfang an
als Lexeme-Parser zu definieren.  So müssen wir das hier nachholen:

> sumOp', mulOp' :: Parser (Int -> Int -> Int)
> sumOp' = lexeme sumOp
> mulOp' = lexeme mulOp
>
> int' :: Parser Int
> int' = lexeme int


Damit können wir Arithmetische Ausdrücke mit Leerzeichen dazwischen
parsen:

> expr :: Parser Int
> expr = flip chainl1 sumOp' . flip chainl1 mulOp' $ int' <|> parens expr


Die ganze Eingabe zwischen führenden Leerzeichen und EOF lesen wir mit

> entirely :: Parser a -> Parser a
> entirely = between spaces eof

    test (entirely expr) " 4 * (2 + 3) "



So, jezt machen wir noch ein Programm draus:

> main :: IO ()
> main
>   = interact
>     $
>     (++"\n") . either show show . parse (entirely expr) "stdin"

    $ ghc calculator.lhs
    $ ./calculator <<< '1 + 2 *3-4'
    3 
          — Enjoy!


Viele interessante Parser-Kombinatoren sind in der Parsec
API-Dokumentation [8] beschrieben...


----------------------------------------------------------------------
Literatur


[1] Daan Leijen.  Parsec, a fast combinator parser.
    http://research.microsoft.com/en-us/um/people/daan/download/parsec/parsec.pdf
[2] http://hackage.haskell.org/package/parsec
[3] https://www.haskell.org/cabal/
[4] https://www.haskell.org/platform/
[5] https://www.archlinux.org/packages/?arch=x86_64&arch=any&q=cabal-install
[6] https://github.com/aslatter/parsec
[7] https://www.stackage.org/haddock/lts-6.2/parsec-3.1.11/Text-Parsec.html#t:ParsecT
[8] http://hackage.haskell.org/package/parsec-3.1.11/docs/Text-Parsec.html#g:2


======================================================================
