
Konzepte der Programmierung                       Universität Konstanz
Programmierkurs 2                Informatik & Informationswissenschaft
Sommer 2016                                Marc Scholl, Stefan Klinger

======================================================================

                             4. Mai 2016

----------------------------------------------------------------------
Organisation


Feiertag
--------

Am Freitag, 6. Mai, werden die üblichen Tutorien angeboten, d.h.:

    D  8:15 F428  Denis Gietz
    E 10:00 F429  Johannes Fuchs vertritt Robert Schmis
    F 11:45 G308  Johannes Fuchs vertritt Fabian Späh

Die Donnerstags-Teilnehmer besuchen bitte *ebenfalls* diese drei
Tutorien, idealerweise bei ihrem jeweiligen Tutor.  Versuchen Sie sich
*gleichmäßig* zu verteilen um Überfüllung zu vermeiden (z.B. A→D, B→E,
und C→F).



Klausurtermine
--------------

Verbindliche Informationen zu den Klausurterminen finden Sie auf den
Fachbereichsseiten:

    https://www.informatik.uni-konstanz.de/studieren/studium/pruefungen/pruefungstermine/


----------------------------------------------------------------------
Anmerkungen zu Aufgabe 3


Viele imperative Sprachen erlauben es, ein `if` ohne `else` zu
schreiben:

    if (x > 42) { print("big x"); }

Das macht Sinn, denn manchmal ist im “anderen Fall” nichts zu tun.


Manche imperative Programmiersprachen (z.B. C, Tiger) erlauben es
außerdem, einen *einzelnen* Befehl *ohne* geschweifte Klammern in
`if`-Klauseln zu verwenden:

    if (x > 42) print("big x");
    
statt

    if (x > 42) { print("big x"); }

Die Klammern sind dann nur nötig, wenn *mehrere* Befehle zu dem
`then`- oder `else`-Zweig gehören:

    if (x > 42) {    
        print("big x");
        x = x - 10;
    }

Genauso verhält es sich mit dem `else`-Zweig:

    if (y < 23) print("small y"); else print("normal y");

Dadurch soll die Sprache lesbarer werden.


Die Grammatik dazu sieht so aus:

    <Stmt> → <Variable> `=` <Expr> `;`
          ...
           | `if` `(` <Expr> `)` <Stmt>                         (1)
           | `if` `(` <Expr> `)` <Stmt> `else` <Stmt>           (2)
          ...
           | `print(` <Expr> `)`

Dabei ist

  <Stmt> das Nonterminal für “statements”, d.h. Anweisungen die
         üblicherweise mit `;` abgeschlossen werden, und irgendetwas
         tun.

  <Expr> das Nonterminal für “expressions”, d.h. Ausdrücke die einen
         Wert liefern.


Ein weiteres Beispiel wäre dieses:

    if (x > 42)
        if (y > 23)
            print("foo");
        else
            print("bar");

Was fällt auf?

Schreiben wir das mal in eine Zeile:

    if (x > 42) if (y > 23) print("foo"); else print("bar");

Zu welchem `if` gehört das `else`?  Und warum?


Diese Problem ist als “Dangling `else` problem” bekannt.  Die
Grammatik lässt beide Parse-Trees zu:

      <Stmt>
      ^^^^^^
    →     | verwenden (1)
    
      if ( <Expr> ) <Stmt>
    →      ^^^^^^
      if ( x > 42 ) <Stmt>
                    ^^^^^^
    →     | verwenden  (2)
                    
      if ( x > 42 ) if ( <Expr> ) <Stmt>        else <Stmt>
    →                    ^^^^^^   ^^^^^^             ^^^^^^
      if ( x > 42 ) if ( y > 23 ) print("foo"); else print("bar");

oder

      <Stmt>
      ^^^^^^
    →     | verwenden (2)
    
      if ( <Expr> ) <Stmt>                      else <Stmt>
    →      ^^^^^^                                    ^^^^^^
      if ( x > 42 ) <Stmt>                      else print("bar");
                    ^^^^^^  
    →     | verwenden (1)
    
      if ( x > 42 ) if ( <Expr> ) <Stmt>        else print("bar");
    →                    ^^^^^^   ^^^^^^        
      if ( x > 42 ) if ( y > 23 ) print("foo"); else print("bar");


Wie kann man dieses Problem lösen?

  * Immer geschweifte Klammern erzwingen.

  * Python: Einrückung wird Teil der Syntax.
 
        if x > 42:                          if x > 42:           
            if y > 23:                          if y > 23:       
                print("foo")       vs.              print("foo") 
            else:                           else:                
                print("bar")                    print("bar")

            ^ zum inneren `if`              ^ zum äußeren `if`

  * Konvention: Bei C gehört das `else` zum letzten vorangehenden `if`
    das noch kein `else` hat [1].

        if (x > 42) if (y > 23) print("foo"); else print("bar");

    wird also als

        if (x > 42) { if (y > 23) print("foo"); else print("bar"); }

    gelesen, und nicht als
    
        if (x > 42) { if (y > 23) print("foo"); } else print("bar");



Haskell
-------

  * In Haskell hat ein `if` immer ein `then` *und* ein `else`!

  * `if`, `then`, und `else` sind Schlüsselworte, die den Code
    lesbarer machen.  Tatsächlich kann man das aber auch selber
    konstruieren:

> foo x = if x > 23
>         then 'a'
>         else 'b'



> wenn :: Bool -> t -> t -> t
> wenn c t e | c         = t
>            | otherwise = e

Ausprobieren:

    wenn True "yes" "no"
    
    wenn False "yes" "no"


Damit kann man ganz normal programmieren:

> f x =  wenn (even x) "gerade" "ungerade"

    f 42
    
    f 23


Frage: Kann man die syntaktische Form für bedingte Ausdrücke auch in
       anderen Sprachen selber konstruieren?


    int wenn(bool c, int t, int f) {

        if (c)
            return t;

        return f;
    }

Das ist valider C-Code.  Welches Problem tritt hier auf?


  * Erstens: Wir verwenden immer noch `if`.

  * eineinhalbtens: Typen.

  * Zweitens: Betrachten `fib` vom 3. Übungsblatt

> fib :: Integer -> Integer
> fib 0 = 0
> fib 1 = 1
> fib n = fib (n-1) + fib (n-2)

Sie haben wahrscheinlich festgestellt dass `fib 31` recht lange dauert.

    fib 31

Was ist mit

    wenn True 23 (fib 31)
    
    wenn False 23 (fib 31)


Der springende Punkt ist die “Lazy” Auswertung in Haskell: `fib 31`
wird nur ausgewertet wenn nötig.

    wenn True 23 (fib 100)
    


----------------------------------------------------------------------
Anmerkungen zu Aufgabe 1


RegEx für gerade natürliche Zahlen in Dezimaldarstellung ohne führende
Null (Leerzeichen zur besseren Lesbarkeit, sind nicht Bestandteil des
Alphabets):

    [02468] | [1-9] [0-9]* [02468]    

Testen an der Shell (cf. Vorlesung “Schlüsselquali Informatik”):

    $ cmp <(seq 0 2 100000) <(seq 0 100000 |
      grep -Ex '[02468]|[1-9][0-9]*[02468]')



Beinahe wäre mir ein Missgeschick passiert:

Idee für eine “nette zweite Teilaufgabe”: Geben sie analog zum ersten
Teil einen Regex für die Dezimaldarstellungen durch 3 teilbarer
Natürlicher Zahlen an.

  * Quersumme muss durch 3 teilbar sein.

  * Interessant: Wir können in Regex nicht rechnen!

Idee: Überlege für jede Ziffer einer Zahl, wie sich die bisherige
      Quersumme ändert.

  * Vereinfachung: Führende Nullen, und die leere Ziffernfolge
    erlauben, d.h., "00093" und "" wären ok.



Lösung:

    ((([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])|[147](([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])*[147](([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])*[258](([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])*[258]|[258](([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])*[258](([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])*[147](([0369]|[147][0369]*[258]|[258][0369]*[147])|[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]([0369]|[147][0369]*[258]|[258][0369]*[147])*[147]|[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258]([0369]|[147][0369]*[258]|[258][0369]*[147])*[258])*[147])*



Wie kommt man auf sowas?

  * Informatik 1 / Theoretische Informatik: Analogie von RegEx und
    Automaten.

                                   _
                          0,3,6,9 / \
                                  | V
                                +-------+     2,5,8 
             +----------------> | Rest  |-------------------+
             |  +---------------| ist 2 | <--------------+  |
             |  |   1,4,7       +-------+                |  |
             |  |                                        |  |
             |  |                                        |  |
       2,5,8 |  |                                  1,4,7 |  |
             |  |                                        |  |
             |  V                                        |  V
          +-------+     1,4,7                          +-------+
          | Rest  |----------------------------------> | Rest  |
          | ist 0 | <----------------------------------| ist 1 |
          +-------+                      2,5,8         +-------+
             | ^                                          | ^
             \_/                                          \_/  
       0,3,6,9                                      0,3,6,9


    Die Zustände des Automaten repräsentieren den Rest der Quersumme
    durch 3.

    Die Kanten des Automaten geben an, wie sich der Zustand beim Lesen
    einer weiteren Ziffer ändert.


Beobachtungen:

  * An jeder Kante steht eine von drei Zeichengruppen, die wir durch
    einen RegEx darstellen können:

        A = [0369]          B = [147]          C = [258]

                                   _
                                A / \       
                                  | V
                                +-------+   C    
             +----------------> | Rest  |-------------------+
             |  +---------------| ist 2 | <--------------+  |
             |  |           B   +-------+                |  |
             |  |                                        |  |
             |  |                                        |  |
             |  |           rechts herum: C              |  |
           C |  |            links herum: B            B |  |
             |  |                                        |  |
             |  V                                        |  V
          +-------+   B                                +-------+
          | Rest  |----------------------------------> | Rest  |
          | ist 0 | <----------------------------------| ist 1 |
          +-------+                                C   +-------+
             | ^                                          | ^
             \_/                                          \_/  
            A                                            A


  * Wenn wir bei dem Zustand “0” anfangen, dann sind wir an Pfaden
    interessiert die den Rest der Quersumme nicht ändern.  So
    bleiben wir bei den durch 3 teilbaren Zahlen.


Frage: Welche Pfade ändern den Rest nicht?

  * Pfade die “auf der Stelle treten”: A

  * Pfade die *genau* einen Schritt gegen, und einen im Uhrzeigersinn
    laufen, oder umgekehrt:

        C B  |  B C

    und zwischen drin beliebig oft auf der Stelle treten:

        B A* C  |  C A* B

    Zusammen mit den bekannten kürzeren Pfaden A geben wir dem einen
    Namen:

        X = A | BA*C | CA*B                                        (X)
    
  * Pfade die *genau* dreimal linksrum, oder rechtsrum laufen:

        B B B | C C C

    Und zwischendrin beliebig oft X:

        B X* B X* B | C X* C X* C

    Zusammen mit den bekannten kürzeren Pfaden X geben wir dem wieder
    einen Namen:

        Y = X | B X* B X* B | C X* C X* C                          (Y)

  * Analog: Pfade die *genau* zweimal nach links und dann zweimal nach
    rechts laufen, oder umgekehrt:

        B B C C  |  C C B B

    Und zwischendrin beliebig oft Y:

        B Y* B Y* C Y* C  |  C Y* C Y* B Y* B

    Zusammen mit den bekannten kürzeren Pfaden Y geben wir dem wieder
    einen Namen:
    
        Z = Y | BY*BY*CY*C | CY*CY*BY*B                            (Z)

  * Das Ganze kann jetzt beliebig oft durchlaufen werden:

        R = Z*

Einsetzen liefert dann den etwas unhandlichen Ausdruck R von oben.



Tatsächlich gibt es eine Methode [2] um deutlich kompaktere Reguläre
Ausdrücke aus Automaten zu generieren.  Das wird dann geradezu
kompakt — Danke Leo:

    ( ([0369]*[258][0369]*[147])*
      [0369]*
      ([147]|[258][0369]*[258])
      ([0369]*[147][0369]*[258])*
      [0369]*
      ([147][0369]*[147]|[258])
    )*
    ([0369]*[258][0369]*[147])*
    [0369]*



Hinweis: Dieser Abschnitt ist natürlich nicht prüfungsrelevant!
         Und Danke an Leo Wörteler, der das zusammengesucht hat.


----------------------------------------------------------------------
Grammatik für Reguläre Ausdrücke


    <RegEx> → <Character>                    — ein einzelner Buchstabe
            | <RegEx> `*`                    — Wiederholung
            | `(` <RegEx> <RegEx> `)`        — Sequenz
            | `(` <RegEx> `|` <RegEx> `)`    — Alternative

   Hier vollständig geklammert.  Konventionen wie in der Vorlesung
   erlauben es Klammern wegzulassen.

Was ist <Character> ?



Nagut: Eine Aufzählung aller Zeichen im Zeichensatz:

    <Character> → `a` | `b` | ... | `z` | `A` | ...
                | `0` | `1` | ...
                | `⇒` | `∀` | ...
                | `☠` | ...
               ...

Oder kürzer:

    <Character> → “any character”



Aber: Ist `*` ∈ <Character> ?



Ja!  Frage:  Was bedeutet  `x*`?

  * Das Zeichen `x` gefolgt vom Zeichen `*`, oder

  * das Zeichen `x`, beliebig oft wiederholt.



Besser:

    <Character> → “any character, except `|`, `*`, `(`, `)`”

Weil die Operator-Zeichen eine besondere Bedeutung haben sollen.



Frage: Wie baut man einen RegEx der die Zeichenfolge `x*y` matcht?



Escaping
--------

Wir brauchen einen Mechanismus um Zeichen von Meta-Zeichen zu
unterscheiden:

Konvention: Durch Voranstellen eines Backslash `\` wird das
            nachfolgende Symbol zu einem normalen Buchstaben.

`x\*y` — Bedeutet also: Ein `x`, gefolgt von einem `*`, gefolgt von
         einem `y`.

`x*y` — Bedeutet also: Beliebig viele `x`, gefolgt von einem `y`.


Die Produktion dazu:

    <Character> → “any character, except `|`, `*`, `(`, `)`”
                | “`\` followed by one of `|`, `*`, `(`, `)`”


Frage: Was fehlt noch?


Der Backslash selber!

    <Character> → “any character, except `|`, `*`, `(`, `)`, `\`”
                | “`\` followed by one of `|`, `*`, `(`, `)`, `\`”


----------------------------------------------------------------------
Bibliografie


[1] Kernighan, Ritchie.  The C Programming Language.  2nd edition,
    p. 56.  Prentice Hall, 1988.  ISBN 0-13-110362-8.

[2] Christoph Neumann.  Converting Deterministic Finite Automata to
    Regular Expressions.  Mar 16, 2005.
    http://www.liacs.nl/~marcello/FI2/DFA_to_RE.pdf


======================================================================

Ausdruck `R` oben erzeugt mit

    $ sed 's=Z=(Y|BY*BY*CY*C|CY*CY*BY*B)=g; s=Y=(X|BX*BX*B|CX*CX*C)=g; s=X=(A|BA*C|CA*B)=g; s=C=[258]=g; s=B=[147]=g; s=A=[0369]=g' <<< 'Z*'


======================================================================
