
Konzepte der Programmierung                       Universität Konstanz
Programmierkurs 2                Informatik & Informationswissenschaft
Sommer 2016                                Marc Scholl, Stefan Klinger

======================================================================

                             23. Mai 2016

----------------------------------------------------------------------
Zu Aufgabe 1, Y-Combinator


Tutorium:

    Y = λx.  (λy. x (y y))  (λy. x (y y))

Die Fixpunkt-Eigenschaft:

    Y f   →*   f (Y f)

damit ist also `Y f` ein Fixpunkt von `f`, weil

    f (Y f)  ≡  Y f


- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


Es gibt eine ganze Familie von Fixpunkt-Kombinatoren, `Y` ist
lediglich ein Vertreter (entdeckt durch Haskell B. Curry).

Weitere Fixpunkt-Kombinatoren:



(a)  Alan Turing's `T` (Theta):

    A = λx y. y (x x y)

    T = A A


Nachweisen der Fixpunkt-Eigenschaft von `T`:

      T f
    =
      A A f
    =
      (λx. λy. y (x x y)) A f
    →
      (λy. y (x x y))[x←A] f

    →     | `y` ist nicht frei in `A`,
          | ⇒ keine α-Konversion nötig.

      (λy. y (A A y)) f
    →
      (y (A A y))[y←f]
    →
      f (A A f)
    =
      f (T f)



(b)  Klop's Fixpunkt-Kombinator `K`:

    L = λa b c d e f g h i j k l m n o p q s t u v w x y z r.
        r( t h i s  i s  a  f i x e d  p o i n t  c o m b i n a t o r )

    K = L L L L L L L L L L L L L L L L L L L L L L L L L L



Bemerkenswert:  Keiner dieser Kombinatoren wird vom GHCi akzeptiert:

 > yc = \x -> (\y-> x (y y)) (\y-> x (y y))

 > theta = a a
 >     where
 >     a = \x y -> y (x x y)

 > klop = l l l l l l l l l l l l l l l l l l l l l l l l l l
 >   where
 >     l = \a b c d e f g h i j k l m n o p q s t u v w x y z r ->
 >         r( t h i s   i s   a   f i x e d
 >            p o i n t   c o m b i n a t o r
 >          )


Wie können wir trozdem eine Funktion `fix` bauen, die für eine
Funktion `f` einen Fixpunkt `fix f` liefert?

     f (fix f)  ≡  fix f

Einfach umdrehen:

> fix f = f (fix f)

Hier sieht man wie wir die *Namensrekursion* von Haskell verwenden:
Der Name `fix` den wir links definieren, wird rechts verwendet.



Die Funktionen vom Übungsblatt:
 
> f :: (Integer -> Integer) -> (Integer -> Integer)
> f g n = if (n == 0) then 1 else n * g (n-1)

> fact :: Integer -> Integer
> fact = fix f
 

----------------------------------------------------------------------
Zu Aufgabe 3, allEqual


> allEqual :: [Integer] -> Bool
> allEqual [] = True
> allEqual (x:xs) = foldr (\y b -> x==y && b) True xs

> allEqual' :: [Integer] -> Bool
> allEqual' [] = True
> allEqual' (x:xs) = foldl (\b y -> x==y && b) True xs

Was ist der wesentliche Unterschied?

Beobachten Speicherverbrauch:

    allEqual $ replicate 20000000 1
    allEqual' $ replicate 20000000 1

Warum?

Erinnerung: Folie 220:

  * Bei `foldr` entscheidet der Operator, ob Rekursion stattfindet,
    und kann vor Ende der Liste abbrechen.

  * `allEqual'` konstruiert dagegen eine sehr tiefe Struktur von
    `foldl`- Aufrufen, die erst nach Erreichen des Listenendes wieder
    abgebaut wird.

Wie lange brauchen wohl die beiden folgenden Aufrufe?

    allEqual [1..10000000]
    allEqual' [1..10000000]

Und diese?

    allEqual [1..]
    allEqual' [1..]

Erinnerung: Folie 220:

  * `foldl` terminiert sicher nicht auf unendlichen Listen!


Am Beispiel mit `foldl`:

    allEqual' [1..]
  ≡
    allEqual' (1:2:3:4:5:…)

  →     | Sei  b # y = 1==y && b

    foldl (#) True (2:3:4:5:…)
  →
    (…((((True # 2) # 3) # 4) # 5) #… 

Das Ergebnis des gesamten Ausdrucks ist das Ergebnis des
“letzten” `#`-Operators in der nicht endenden Liste.  Den
werden wir nicht finden!

Das wird klarer wenn man sich den AST anschaut:
               
               …
               @
              / \
             @   4
            / \
           #   @
              / \
             @   3
            / \
           #   @
              / \
             @   2
            / \
           #   True
    

Die Variante mit `foldr'`:

    allEqual [1..]
  ≡
    allEqual (1:2:3:4:5:…)
    
  →     | Sei  y # b -> 1==y && b
  
    foldr (#) True (2:3:4:5:…)
  →
    2 # (3 # (4 # (5 #… )…))

  →     | Hier ist die erste `#` auch die Wurzel des Ausdrucks:
        |     
        |         @___
        |        /    \
        |       @      @___    
        |      / \    /    \    
        |     #   2  @      @___ 
        |           / \    /    \
        |          #   3  @      @
        |                / \    / \  
        |               #   4  @   …
        |                     / \  
        |                    #   5 

    1==2 && (3 # (4 # (5 #… )…))
  →
    False && (3 # (4 # (5 #… )…))
  →
    False


Ist `foldr` also immer besser als `foldl`?

  * Sehr oft schon.  Später (PK2 13) werden wir das nochmal genauer
    anschauen.

  * Abschätzung von Laufzeit und Speicherbedarf sind trickreich, und
    meist nicht Teil der Übung.

  ⇒ Versuchen Sie nicht Ihre Programme super-effizient zu machen!
    Achten Sie darauf dass Programme überhaupt terminieren,
    bzw. unendliche Listen *produzieren*.  (Test z.B. mit `take 10`).

    
----------------------------------------------------------------------
Zu Aufgabe 4


Tutorial:

> gaps :: Double -> [Double] -> [(Double, Double)]
> 
> gaps d (x:xs) = f x xs
>     where
>     f x (y:ys) | abs (x-y) > d = (x,y) : f y ys
>                | otherwise     = f y ys
>     f _ _ = []
>     
> gaps _ [] = []


Testen mit einer Beispielliste:

> bsp :: [Double]
> bsp = [1.2, 3.4, 12.5, 6.14, 8, 11]

    gaps 2.2 bsp



Alternativer Plan:

  * Bestimme alle aufeinanderfolgenden Paare.

  * Filtere diejenigen die weit genug auseinander liegen.

> pairs' :: [a] -> [(a, a)]
> 
> pairs' (x:xs) = f x xs
>     where
>     f x (y:ys) = (x,y) : f y ys
>     f _ _ = []
>     
> pairs' _ = []

Testen:

    pairs' bsp

Das kann man noch als Einzeiler formulieren.  Wie?

> pairs :: [a] -> [(a, a)]
> 
> pairs xs = zip xs $ tail xs

Kann jemand die Idee dazu beschreiben?


Jetzt: Filtern Paare mit großem Abstand:

> gaps1 :: Double -> [Double] -> [(Double, Double)]
> 
> gaps1 d xs = filter (\(a,b) -> abs (a-b) > d) $ pairs xs

η-Konversion liefert:

> gaps2 :: Double -> [Double] -> [(Double, Double)]
> 
> gaps2 d = filter (\(a,b) -> abs (a-b) > d) . pairs


Auch schön:

> gaps3 :: Double -> [Double] -> [(Double, Double)]
> 
> gaps3 d = filter ((>d) . abs . uncurry (-)) . pairs

Was tut `uncurry` ?


Hier die Typen:

    uncurry ::  (α → β → c)  →  (α, β) → c

      curry :: ((α, β) → c)  →   α → β → c


----------------------------------------------------------------------
Spielen mit `.` und `$`


Was ist das?

    foldr ($)

    *Main> :t foldr ($)
    foldr ($) :: Foldable t => a -> t (a -> a) -> a

`Foldable` — das lesen wir wie?

    foldr ($) :: a -> [a -> a] -> a

Was tut das wohl?

Testen:

    foldr ($) 7 [(/2), (*10), sqrt, (+2)]



Ich hätte lieber einen Typ wie diesen:

> apply :: [a -> a] -> a -> a

Das schreit geradezu nach `flip`.  Aber `flip ‹was›`?

    flip foldr ($)                                  (1)
    
    flip $ foldr ($)                                (2)
    
    flip . foldr ($)                                (3)

(1) vertauscht die Argumente von `foldr`, `$` wird also zum
    Rückgabewert für die leere Liste.

        foldr :: (a -> b -> b) -> b -> [a] -> b

(2) richtig.  Wir wollen die Argumente von `foldr ($)` vertauschen,
    also

> apply = flip $ foldr ($)

    apply [(/2), (*10), sqrt, (+2)] 7



Ausblick: Typinferenz (später in der Vorlesung)

(3) wendet `flip` nach `foldr ($) an.

        foldr ($) :: a -> [a -> a] -> a

    Also muss

        foldr ($) (irgendwas::a) :: [a -> a] -> a

    eine Funktion liefern, die wir flippen können.

        flip :: (b -> c -> d) -> c -> b -> d

    Also muss der Typ `[a -> a] -> a` dem Typ `b -> c -> d` entsprechen.

        [a -> a] -> a  =  b -> (c -> d)
      ⇒
        b  =  [a -> a]   ∧   a  =  c -> d

    Zusammengesetzt ist

        flip . foldr ($) :: (c -> d) -> c -> [(c -> d) -> c -> d] -> d


- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Geht `zip` nur mit Paaren?


    zip :: [a] -> [b] -> [(a, b)]

    zipWith :: (a -> b -> c) -> [a] -> [b] -> [c]

Bereits vordefiniert sind ebenfalls:

    zip3 :: [a] -> [b] -> [c] -> [(a, b, c)]
    
    zipWith3 :: (a -> b -> c -> d) -> [a] -> [b] -> [c] -> [d]

Was tun die wohl?

Testen:

    zipWith3 (\a b c-> if c then a * b else a + b) [1..] [2..] [True, False]



Wie bauen wir uns ein `zip4`?



Plan A
------

> zip4 :: [a] -> [b] -> [c] -> [d] -> [(a, b, c, d)]
> zip4 = zipWith4 (,,,)

Welchen Typ müsste `zipWith4` haben?


> zipWith4 :: (a -> b -> c -> d -> t) -> [a] -> [b] -> [c] -> [d] -> [t]

Konstruktion?

> zipWith4 f = go
>   where
>     go (w:ws) (x:xs) (y:ys) (z:zs) = f w x y z : go ws xs ys zs
>     go _      _      _      _      = []

zip4 [1..] ['a'..] [True, False, False] [10, 20..]


Dieser Ansatz ist etwas unbefriedigend.  Warum?


Wir müssten für jedes `n` eine Funktion `zipWith‹n›` konstrueiren.


Plan B         (kommt später nochmal, cf. [1])
------

> zapp :: [b -> c] -> [b] -> [c]
> zapp = zipWith ($)

Was tut `zapp`?




Testen:

    zapp [(+1), (*2), (/3)] [10, 20..]


Fest eingebaut ist `repeat :: a -> [a]`.  Was tut das wohl?

    repeat 42

Was passiert hier:

    repeat (*2) `zapp` [1..10]

    repeat (+) `zapp` [1..10] `zapp` [100, 200..]


> infixl 0 <>
> (<>) = zapp

    repeat (+) <> [1..10] <> [100, 200..]

Wie kann man damit die 4 Listen

    [1..]
    ['a'..]
    [True, False, False]
    [10, 20..]

in eine Liste von Quadrupeln “zippen”?

    repeat (,,,) <> [1..] <> ['a'..] <> [True, False, False] <> [10, 20..]


____________________
[1] Conor McBride, Ross Paterson.  Functional Pearl, Idioms:
    Applicative programming with effects.
    <http://strictlypositive.org/Idiom.pdf>



======================================================================
